const DEFAULT_TIMEOUT = 60000*5;

export class CacheEntry {
  public key: string;
  public element: any;
  public time: number;
  public timeout?: number;

  constructor(key: string, element: any, time: number, timeout?: number) {
    this.key = key;
    this.element = element;
    this.time = time;
    this.timeout = timeout;
  }  

}

/**
 * Memory-based cache with automatic, time-based key invalidation.
 */
export class MemCache { 
  private static _cache: any = {};
  private static _timeout = DEFAULT_TIMEOUT;

  private static setTimeout(timeout: number) {
    MemCache._timeout = timeout;
  }

  public static add(key: string, element: any, timeout?: number) {
    MemCache._cache[key] = new CacheEntry(key, element, Date.now(), timeout);
  }

  public static remove(key: string) {
    delete MemCache._cache[key];
  }

  public static retrieve(key: string) {
    let entry = MemCache._cache[key];
    // cache entry is invalid
    if (!MemCache.isValid(key)) {
      delete MemCache._cache[key];
      return null;
    } else if (entry) {
      return entry.element;
    }
  }

  /**
   * Returns true if the entry with key is in the cache and is not expired.
   * @param key a key of the cache
   */
  public static isValid(key: string):boolean {
    return MemCache._cache[key] && !MemCache.isExpired(key);
  }

  /**
   * Returns true iif the entry associated with key has been added longer than the cache's 
   * timeout and item's timeout.
   * @param key a valid key of the cache
   */
  public static isExpired(key: string): boolean {
    return MemCache._cache[key].timeout && MemCache._cache[key].time < Date.now() - MemCache.getTimeout(MemCache._cache[key]);
  }

  public static getTimeout(entry: CacheEntry): number {
    return entry.timeout || MemCache._timeout;
  }
}

