class TimedAction {
    public key : string;
    public element : object;
    public action : Function;
    public timeoutId : string;
    public promise : Promise<any>;

    constructor(key : string, element : object, action : Function, timeoutId : string, promise : Promise<any>) {
        this.key = key
        this.element = element
        this.action = action
        this.timeoutId = timeoutId,
        this.promise = promise
    }
  }
  
  /**
   * Implements a time-sensitive, key-based function execution pipeline.
   */
  export default class TimedActionManager {
    static _instance : TimedActionManager;
    
    private timers : any;
    private actions : any;

    constructor() {
        
        this.timers = {};
        this.actions = {};
    }

    public static get instance() : TimedActionManager {
        if (!TimedActionManager._instance) {
            TimedActionManager._instance = new TimedActionManager();
        }
        return TimedActionManager._instance;
    }
  
    // public functions
  
    /**
     * Schedules action to be executed on element after delay milliseconds, or updates the element if it's already in place
     * 
     * @param {String} key 
     * @param {Object} element 
     * @param {Function} action 
     * @param {Number} delay 
     */
    start (key : string, element : object, action : Function, delay : number): Promise<any> {
        if (!this.hasAction(key)) {
            return this.schedule(key, element, action, delay)
        } else {
            return this.merge(key, element)
        }
    }
  
    /**
     * Cancels the execution of a timed action
     * @param {String} key The scheduled action's key
     */
    cancel(key : string): TimedActionManager {
        if (this.hasAction(key)) {
            clearTimeout(this.actions[key].timeoutId)
            // TODO: resolve the promise
            delete this.actions[key]
        }
        return this
    }
  
    /**
     * Immediately executes action over element, even if it was already scheduled.
     * @param {String} key The key for our action
     * @param {Object} element Element to be processed by the action
     * @param {Function} action function to be applied 
     */
    immediate(key : string, element : object, action : Function) {
        return this.start(key, element, action, 0)
    }
  
    // private functions
    
    /**
     * Checks if there's already an item with the given key
     * @param {String} key an element's key
     */
    private hasAction(key : string): boolean {
        return !!this.actions[key]
    }
  
    private schedule(key : string, element : object, action : Function, delay : number): Promise<any> {
      let me = this
      let timeoutId = null
      let myPromise = new Promise((resolve, reject) => {
        
        timeoutId = setTimeout(() => {
          me.execute(key).then((result : any) => {
            resolve(result);
          }).catch((error : any) => {
            reject(error);
          });
        }, delay);
        
      });
      me.addToQueue(key, element, action, timeoutId, myPromise)
      return myPromise
    }
  
    private addToQueue(key : string, element : object, action : Function, timeoutId : any, promise : Promise<any>): TimedActionManager {
      this.actions[key] = new TimedAction(key, element, action, timeoutId, promise);
      return this;
    }
  
    private merge(key : string, element : object): Promise<any> {
      let existing = this.actions[key];
      existing.element = Object.assign(existing.element, element);
      return existing.promise;
    }
  
    private execute(key : string): Promise<any> {
      if (this.hasAction(key)) {
          let timedAction = this.actions[key]
          delete this.actions[key]
          return timedAction.action(timedAction.key, timedAction.element)
      }
      return new Promise((resolve, reject) => {
        reject(null)
      })
    }
  
  }